(* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *)
(*  Soubor: PALETA.PAS                                                     *)
(*  Obsah: procedury pro operace s paletou                                 *)
(*  Autor: Mircosoft (vetsinou)                                            *)
(*  Posledni uprava: 26.7.2005                                             *)
(*  Pro kompilaci: CAS.PAS/TPU, GLOBAL.PAS/TPU                             *)
(*  Pro spusteni: Pokud chcete pracovat se 16tibarevnymi paletami, musi    *)
(*                byt v jednotce Global nastaven spravny kod grafickeho    *)
(*                rezimu, aby se rozlisily 16ti a 256tibarevne rezimy.     *)
(*                (jine indexy barev) O to by se mela postarat jednotka,   *)
(*                ktera grafiku spousti (VGAstart, SVGA640, SVGA800...).   *)
(*  Upozorneni: tyto zdrojove kody pouzivate na vlastni nebezpeci          *)
(* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *)
unit Paleta;
{$G+}
interface

{Vsechny parametry oznacene Rychlost znamenaji dobu cekani mezi dvema kroky
v ms (pro proceduru Pockej). Pocet kroku, ktere se provedou, neni vzdy stejny,
zavisi na tom, o kolik se lisi aktualni paleta od te pozadovane (maximalne
tedy 63 kroku). Optimalni hodnota rychlosti pro caste pouzivani je neco mezi
1 a 10, 50 a vic uz je opravdu hodne pomale. Kdyz zadate nulu, zadne plynule
prechody nebudou a cela akce se provede skokove v jednom kroku.
Cekani je reseno tak, aby nezaviselo na rychlosti procesoru.

Jas je v procentech, jde nastavit od 0% (cerna obrazovka) az do 200% (uplne
bila obrazovka), normalni je 100%. Nastavovani jasu ovlivnuje jen zobrazeni
barev na obrazovce, na hodnotach v palete (promenne) se neprojevuje.
Pri nacteni palety z obrazovky se ale vsechny barvy nactou "natvrdo",
tj. vcetne aktualniho zjasneni nebo ztmaveni.}


(****************** 16 *******************************************************
Univerzalni procedury pro 16tibarevne palety, funguji ve vsech 16tibarevnych
rezimech vcetne textoveho i ve 256tibarevnych (pak se manipuluje jen s prvnimi
16ti barvami): *)

type RGBpal16=array[0..15,{cisla barev}
                    1..3]{cisla slozek - 1=cervena, 2=zelena, 3=modra}
                     of byte;{hodnoty slozek jsou od 0 do 63}

procedure DefaultPal(var P:rgbpal16);
{Palete P priradi standardni barvy. Nic nezobrazuje ani nenacita z obrazovky,
jen priradi par konstantnich hodnot.}
procedure ZjistiPaletu(var P:rgbpal16);
{do promenne P ulozi aktualni paletu z obrazovky}
procedure NastavPaletu(var P:rgbpal16;NovyJas:byte);
{Nastavi na obrazovce paletu P s prislusnym jasem. P je VAR jenom proto, aby
se zbytecne nevytvarela lokalni promenna, ale jinak se s ni nic nedela
(je to rychlejsi).}
procedure NastavBarvu(var P:rgbpal16;Cislo,R,G,B:byte);
{nastavi barvu Cislo v palete P i na obrazovce, jas se nemeni}
procedure ZmenJas(var P:rgbpal16;NovyJas:byte;Rychlost:word);
{plynula zmena jasu vsech barev}
procedure ZmenBarvu(var P:rgbpal16;CisloBarvy,nR,nG,nB:byte;Rychlost:word);
{plynule prebarveni jedne barvy na jinou, jas zustava konstantni}
procedure ZmenPaletu(Start:rgbpal16;var Cil:rgbpal16;Rychlost:word);
{plynule prebarveni z palety Start do palety Cil. Paleta Start uz neni VAR,
protoze ji potrebuju jako lokalni promennou (aby se nemenila).}

(******************* 256 *****************************************************
procedury jen pro 256tibarevne rezimy:  *)

type RGBpal256=array[0..255,1..3]of byte;

procedure _ZjistiPaletu(var Pal:rgbpal256);
{nacte celou paletu z obrazovky a ulozi ji do promenne Pal}
procedure _NastavCelouPaletu(var Pal:rgbpal256);
{aplikuje paletu Pal. Velmi rychla (100% asm), ale nepracuje s jasem}
procedure _NastavPaletu(var P:rgbpal256;Vocad,Pocad,NovyJas:byte);
{nastavi barvy s cisly Vocad az Pocad z palety P s jasem Novyjas}
procedure _NastavBarvu(var P:rgbpal256;Cislo,R,G,B:byte);
{nastavi jednu barvu v palete i na obrazovce, jas se nemeni}
procedure _ZmenJas(var P:rgbpal256;Vocad,Pocad,NovyJas:byte;Rychlost:word);
{plynula zmena jasu, jen pro barvy Vocad..Pocad}
procedure _ZmenBarvu(var P:rgbpal256;CisloBarvy,nR,nG,nB:byte;Rychlost:word);
{plynule prebarvi jednu barvu na jinou (v palete i na monitoru)}
procedure _ZmenPaletu(Start:rgbpal256;var Cil:rgbpal256;Vocad,Pocad:byte;Rychlost:word);
{plynule prebarveni barev vocad..pocad z palety Start do Cil, jas zustava}
procedure _RotujPaletu(P:rgbpal256;vocad,pocad:byte;oKolik:integer);
{posune dany usek barev o oKolik pozic (okolik>0 => doprava, oKolik<0 => do-
leva) a ty barvy, ktere "vypadnou" jednim koncem ven, vrati druhym koncem
zpatky. Jas se nemeni. Efekt se projevi jen na obrazovce, v palete ne.}

implementation
uses cas,global;

const jas:byte=100;{0 = tma, 100 = normal, 200 = bilo}

var sestnact:boolean;{pri inicializaci se sem ulozi, jestli jsme v 16ti nebo
256ti barevnem rezimu. Zjisti se to z jednotky Global, ktere to zase rekla
jednotka, ktera ma start grafiky na svedomi. Z cehoz vyplyva: do Uses piste
jednotku Paleta az ZA jednotku pro spusteni grafiky!
Pokud spoustite grafiku jinak, musite si tuhle promennou nastavit rucne.
Jestli univerzalni 16tibarevne procedury nechcete pouzivat, muzete se na ni
uplne vykaslat, smazat cele propojeni s jednotkou Global atd..}

procedure SetRGBpal(cislo,r,g,b:byte); assembler;
{rychlejsi a univerzalnejsi obdoba Setrgbpalette, pouziti uplne stejne}
Asm
mov dx, 03C8h        {coz znamena:}
mov al, cislo
out dx, al           {port[$3C8]:=cislo;}
inc dx
mov al, R
out dx, al           {port[$3C9]:=r;}
mov al, G
out dx, al           {port[$3C9]:=g;}
mov al, B
out dx, al           {port[$3C9]:=b;}
End;{setrgbpal}

Procedure getrgbpal(cislo:byte;var R,G,B:byte); assembler;
{nacteni barvy}
Asm
mov dx, $03C7
mov al, cislo
out dx, al               {port[$037C]:=cislo;}
inc dx
inc dx
in al, dx
les di, R
mov es:[di], al          {R:=port[$03C9];}
in al,dx
les di, G
mov es:[di], al          {G:=port[$03C9];}
in al,dx
les di, B
mov es:[di], al          {B:=port[$03C9];}
End;{getrgbpal}

(************************ 16 ************************************************)

function prevod(cislo:byte):byte;{tohle zajistuje tu univerzalnost}
const pole:array[0..15]of byte=(0,1,2,3,4,5,20,7,56,57,58,59,60,61,62,63);
Begin
if sestnact then prevod:=pole[cislo]{skutecny index barvy v 16tibarevnem rezimu neni 0..15, ale tohle}
            else prevod:=cislo;{ve 256 barvach neni problem}
End;{prevod}

procedure defaultpal(var p:rgbpal16);
const defaultpalette:rgbpal16=(( 0, 0, 0),( 0, 0,42),( 0,42, 0),( 0,42,42),
(42, 0, 0),(42, 0,42),(42,21, 0),(42,42,42),(21,21,21),(21,21,63),(21,63,21),
(21,63,63),(63,21,21),(63,21,63),(63,63,21),(63,63,63));{vychozi paleta, jaka je bezne nastavena pri spusteni grafiky}
Begin
p:=defaultpalette;
End;{defaultpal}

procedure ZjistiPaletu(var P:rgbpal16);
var i:byte;
Begin
for i:=0 to 15 do getrgbpal(prevod(i),p[i,1],p[i,2],p[i,3]);
End;{zjistipaletu}

procedure nastavpaletu(var p:rgbpal16;novyjas:byte);
var i:byte;
Begin
jas:=novyjas;
if jas>100 then for i:=0 to 15 do setrgbpal(prevod(i),p[i,1]+((63-p[i,1])*(jas-100))div 100,{zjasneni}
                                                      p[i,2]+((63-p[i,2])*(jas-100))div 100,
                                                      p[i,3]+((63-p[i,3])*(jas-100))div 100)
           else for i:=0 to 15 do setrgbpal(prevod(i),(p[i,1]*jas)div 100,{stmivani}
                                                      (p[i,2]*jas)div 100,
                                                      (p[i,3]*jas)div 100);
End;{nastavpaletu}

procedure nastavbarvu(var P:rgbpal16;cislo,r,g,b:byte);
Begin
p[cislo,1]:=r; p[cislo,2]:=g; p[cislo,3]:=b;
if jas>100 then setrgbpal(prevod(cislo),r+((63-r)*(jas-100))div 100,g+((63-g)*(jas-100))div 100,b+((63-b)*(jas-100))div 100)
           else setrgbpal(prevod(cislo),(r*jas) div 100,(g*jas) div 100,(b*jas) div 100);
End;{nastavbarvu}

procedure zmenjas(var p:rgbpal16;novyjas:byte;rychlost:word);
Begin
if rychlost=0 then begin{skokem-pro pomale pocitace nebo netrpelive uzivatele}
                   jas:=novyjas;
                   nastavpaletu(p,jas);
                   end
              else begin{plynule}
                   while jas<>novyjas do begin
                                         startcekani;
                                         if jas>novyjas then dec(jas)
                                                        else inc(jas);
                                         nastavpaletu(p,jas);
                                         pockej(rychlost);
                                         end;
                   end;
End;{zmenjas}

procedure ZmenBarvu(var P:rgbpal16;cislobarvy,nr,ng,nb:byte;rychlost:word);
var r,g,b:byte;
Begin
if rychlost=0 then begin
                   nastavbarvu(p,cislobarvy,nr,ng,nb);
                   end
              else begin
                   r:=p[cislobarvy,1];
                   g:=p[cislobarvy,2];
                   b:=p[cislobarvy,3];
                   while(r<>nr)or(g<>ng)or(b<>nb)do
                     begin
                     startcekani;
                     if r>nr then dec(r)else if r<nr then inc(r);
                     if g>ng then dec(g)else if g<ng then inc(g);
                     if b>nb then dec(b)else if b<nb then inc(b);
                     nastavbarvu(p,cislobarvy,r,g,b);
                     pockej(rychlost);
                     end;
                   end;
End;{zmenbarvu}

procedure ZmenPaletu(start:rgbpal16;var cil:rgbpal16;rychlost:word);
var i,j:byte;
    padla:boolean;
{}function hotovo:boolean;{protoze nejde napsat if start=cil then..., musim na to takhle}
{}var it:byte;
{}Begin
{}it:=0;
{} repeat{z forcyklu bych musel vyskakovat Breakem nebo ho projet az do konce. Oboji se mi nelibi}
{} if(start[it,1]<>cil[it,1])or(start[it,2]<>cil[it,2])or(start[it,3]<>cil[it,3])
{}  then it:=20;
{} inc(it);
{} until it>15;
{}hotovo:=(it=16);
{}End;{hotovo}
Begin{zmenpaletu}
padla:=false;
if rychlost=0 then nastavpaletu(cil,jas)
              else repeat
                   startcekani;
                   if hotovo then padla:=true{hrozna konstrukce,ale potrebuju volat pomalou funkci Hotovo az po Startcekani}
                             else begin{paletu Start pouzivam jako lokalni promennou}
                                  for i:=0 to 15 do
                                   for j:=1 to 3 do if start[i,j]>cil[i,j] then dec(start[i,j])
                                                                           else if start[i,j]<cil[i,j] then inc(start[i,j]);
                                  nastavpaletu(start,jas);
                                  end;
                   pockej(rychlost);
                   until padla;
End;{zmenpaletu}

(************************** 256 *********************************************)

{setrgbpal a getrgbpal funguji i tady}

procedure _NastavCelouPaletu(var pal:rgbpal256); assembler; {by Asp / VR group}
Asm
mov dx,$03C8
xor al,al
out dx,al    {port[$03C8]:=0}
push ds
lds si,pal   {priprav do DS:SI adresu palety}
{cld}
inc dx
mov cx,768
rep outsb    {port[$03C9]:=postupne kazdy byte palety}
pop ds
End;{_nastavceloupaletu}

procedure _ZjistiPaletu(var pal:rgbpal256); assembler; {by Asp / VR group}
Asm
mov dx,$03C7
xor al,al
out dx,al
les di,pal
{cld}
inc dx
inc dx
mov cx,768
rep insb
End;{_nacticeloupaletu}

procedure _NastavPaletu(var p:rgbpal256;vocad,pocad,novyjas:byte);
var i:byte;
Begin
jas:=novyjas;
if jas>100 then for i:=vocad to pocad do setrgbpal(i,p[i,1]+((63-p[i,1])*(jas-100))div 100,
                                                     p[i,2]+((63-p[i,2])*(jas-100))div 100,
                                                     p[i,3]+((63-p[i,3])*(jas-100))div 100)
           else for i:=vocad to pocad do setrgbpal(i,(p[i,1]*jas)div 100,
                                                     (p[i,2]*jas)div 100,
                                                     (p[i,3]*jas)div 100);
End;{_nastavpaletu}

procedure _nastavbarvu(var p:rgbpal256;cislo,r,g,b:byte);
Begin
p[cislo,1]:=r; p[cislo,2]:=g; p[cislo,3]:=b;
if jas>100 then setrgbpal(cislo,r+((63-r)*(jas-100))div 100,g+((63-g)*(jas-100))div 100,b+((63-b)*(jas-100))div 100)
           else setrgbpal(cislo,(r*jas) div 100,(g*jas) div 100,(b*jas) div 100);
End;{_nastavbarvu}

procedure _zmenjas(var p:rgbpal256;vocad,pocad,novyjas:byte;rychlost:word);
Begin
if rychlost=0 then begin
                   jas:=novyjas;
                   _nastavpaletu(p,vocad,pocad,jas);
                   end
              else begin
                   while jas<>novyjas do begin
                                         startcekani;
                                         if jas>novyjas then dec(jas)
                                                        else inc(jas);
                                         _nastavpaletu(p,vocad,pocad,jas);
                                         pockej(rychlost);
                                         end;
                   end;
End;{_zmenjas}

procedure _ZmenBarvu(var P:rgbpal256;cislobarvy,nr,ng,nb:byte;rychlost:word);
var r,g,b:byte;
Begin
if rychlost=0 then begin
                   _nastavbarvu(p,cislobarvy,nr,ng,nb);
                   end
              else begin
                   r:=p[cislobarvy,1];
                   g:=p[cislobarvy,2];
                   b:=p[cislobarvy,3];
                   while(r<>nr)or(g<>ng)or(b<>nb)do
                     begin
                     startcekani;
                     if r>nr then dec(r)else if r<nr then inc(r);
                     if g>ng then dec(g)else if g<ng then inc(g);
                     if b>nb then dec(b)else if b<nb then inc(b);
                     _nastavbarvu(p,cislobarvy,r,g,b);
                     pockej(rychlost);
                     end;
                   end;
End;{_zmenbarvu}

procedure _ZmenPaletu(start:rgbpal256;var cil:rgbpal256;vocad,pocad:byte;rychlost:word);
var i,j:byte;
    padla:boolean;
{}function hotovo:boolean;
{}var it:word;
{}Begin
{}it:=0;
{} repeat
{} if(start[it,1]<>cil[it,1])or(start[it,2]<>cil[it,2])or(start[it,3]<>cil[it,3])
{}  then it:=257;
{} inc(it);
{} until it>255;
{}hotovo:=(it=256);
{}End;{hotovo}
Begin{_zmenpaletu}
padla:=false;
if rychlost=0 then _nastavpaletu(cil,vocad,pocad,jas)
              else repeat
                   startcekani;
                   if hotovo then padla:=true
                             else begin
                                  for i:=vocad to pocad do
                                   for j:=1 to 3 do if start[i,j]>cil[i,j] then dec(start[i,j])
                                                                           else if start[i,j]<cil[i,j] then inc(start[i,j]);
                                  _nastavpaletu(start,vocad,pocad,jas);
                                  end;
                   pockej(rychlost);
                   until padla;
End;{_zmenpaletu}

procedure _RotujPaletu(P:rgbpal256;vocad,pocad:byte;oKolik:integer);
var barva:byte;
    i:word;
    pom:array[1..3]of byte;
Begin
if okolik=0 then exit;
if okolik>0 then for i:=1 to okolik do begin
                                       move(p[pocad,1],pom[1],3);
                                       for barva:=pocad downto vocad+1 do p[barva]:=p[barva-1];
                                       move(pom[1],p[vocad,1],3);
                                       end
            else for i:=1 to -okolik do begin
                                        move(p[vocad,1],pom[1],3);
                                        for barva:=vocad to pocad-1 do p[barva]:=p[barva+1];
                                        move(pom[1],p[pocad,1],3);
                                        end;
_nastavpaletu(p,vocad,pocad,jas);
End;{_rotujpaletu}

BEGIN
sestnact:=(global_grafrezim shr 8=gdVGA)or(global_grafrezim shr 8=gdText);
{nepredpokladam pouziti grafiky VESA16.BGI, ktera ma stejny kod jako 256tibarevna SVGA256.BGI}
END.
